!***********************************************************************
! Module: ommosaic_mod
!
! Authors:  Christopher Brindle - SSC (MSSL)
!          
! Description: Mosaics OM images
! 
! Comments: Based on emosiac_mod.f90 written by R. West - SSC (Leicester)
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!
! Change History
!          July 9 2001 - CB Added check to see if any images completely 
!          overlap another one- if so the program outputs a warning message.
!          November 13 2001 CB - Ammended code to mosiac both low and high-
!          resolution images properly. Also modified to work if input files
!          don't have an exposure map. Modified to produce an exposure map
!          if required.
!          
!************************************************************************ 
module ommosaic_mod

  use dal
  use param  
  use errorHandling
  use constants
  use omnumerics_mod
  use sla

  implicit none 
  
    private   

    !*********************************************
    ! A structure for storing the image parameters
    !*********************************************
    type ImageT
        character (50) :: filename
        real (single), dimension(:, :), pointer :: image
        real (single), dimension(:, :), pointer :: weight 
        real (double) :: xrpix, yrpix, xrval, yrval
        real (double) :: xdelt, ydelt, roll
        real(single) :: exposure
        integer :: datatype, dims(0 : 1), xmin, xmax, ymin, ymax
        integer(int32) :: binAx1
        real (double) :: sky2img(3, 3)
    end type ImageT

    !***************************************
    ! The following structure will store 
    ! the observation details.
    !**************************************
    type OBSERVATIONDETAILS 
        character(40) telescope 
        character(40) instrument      
        character(40) observer      
        character(40) object
        character(40) filter
        character(40) obsID
        character(40) expID
        character(40) startObs
        character(40) endObs
        real(single) exposure
    end type OBSERVATIONDETAILS
    
    public ommosaic

   
contains 
!********************************************
! Subroutine ommosaic
!  
!********************************************
subroutine ommosaic

    type (ImageT), allocatable, dimension(:) :: input
    type (ImageT) :: output
   
    integer i, noFiles
    character(100), allocatable, dimension(:) :: fileName
    character (100) :: outputFile
    integer :: maxIndex
    integer(int32), pointer, dimension(:) :: index
    real (single), dimension(:, :), pointer :: exposureImage
  
    real(single) :: maxPixelSize   
    integer(int32) :: exposureTime    
    integer(int32) :: noHighResImages
    logical :: exposureMap
    type (OBSERVATIONDETAILS) :: obsDetails 

    noFiles = parameterCount("imagesets")

    allocate(fileName(1 : noFiles))
     

    outputFile = stringParameter("mosaicedset")
 
    exposureMap = booleanParameter("exposuremap")   

    exposureTime = realParameter("exposure")
    !****************************************************       
    ! Open input files, read data arrays and WCS keywords
    !****************************************************
    
    allocate(input(1 : noFiles))
    allocate(index(1 : noFiles))
    maxPixelSize = 0.0
    maxIndex = 1
    noHighResImages = 0
     
    do i = 1, noFiles
        fileName(i) = trim(stringParameter("imagesets", i - 1))

        call readImage(input(i), fileName(i), obsDetails)

        !************************************************
        ! Check to see if image is a high-resolution one
        !************************************************
        if (input(i)%binAx1 .eq. 0) then
            noHighResImages = noHighResImages + 1
            index(noHighResImages) = i
        end if
    end do

    if (noHighResImages .gt. 0 .and. noHighResImages .lt. noFiles) then
        !*********************************************
        ! Convert these images to low-resolution ones
        !********************************************* 
        do i = 1, noHighResImages
            call ConvertToLowResImage(input(index(i)))
        end do 
    end if
    deallocate(index)


    do i = 1, noFiles
        !************************************************************
        ! Need to normalise the images to the same exposure time so 
        ! that they can be correctly coadded
        !************************************************************ 
        where(input(i)%image .gt. 0.0) input(i)%image = input(i)%image / input(i)%exposure

        !****************************************
        ! Check the pixel size
        !***************************************
        if (input(i)%xdelt .gt. maxPixelSize) then
            maxIndex = i
            maxPixelSize = input(i)%xdelt
        end if    
    end do           

    !******************************************
    ! Check that the filters are all the same
    !******************************************
    call CheckFilters(noFiles, fileName, obsDetails)


     !******************************************
    ! Check that there are no overlapping windows
    !******************************************
    call CheckWindows(noFiles, fileName)     


    !**************************************************************
    ! For now assume output image has same reference coordinate and
    ! pixel size as the first input image
    !*************************************************************
    output%xrval = input(maxIndex)%xrval
    output%xdelt = input(maxIndex)%xdelt
    output%yrval = input(maxIndex)%yrval
    output%ydelt = input(maxIndex)%ydelt
    output%roll  = input(maxIndex)%roll
       
    !*******************************************************
    ! Calculate size of output image (physical and in pixels)
    !*******************************************************
    call calculateOutputSize(noFiles, input, output)

    !************************************************                                                                          ! Loop through input images adding to the output
    !***********************************************

    allocate(exposureImage(0 : output%dims(0) - 1, 0 : output%dims(1) - 1))
    exposureImage = 0.0
  

    do i = 1, noFiles

        write(errstr,'(a, a, a)') "Adding image in file ", trim(fileName(i)), " to the mosaiced image"
      
        call message(AppMsg, Verbose)

        call addImage(input(i), output, exposureImage)
    end do
    deallocate(input)       
 
    !*****************************************************
    ! Normalise each pixel image by the sum of the 
    ! exposure times for each pixel
    !*****************************************************
    where (output%weight .gt. 0.0) output%image = output%image / real(output%weight)

    output%weight = exposureImage
    deallocate(exposureImage)

      
    where (output%image .gt. 0.0)  output%image = output%image * real(exposureTime)

    obsDetails%exposure = exposureTime
 
    call writeImage(output, outputFile, obsDetails, exposureMap)        


    !******************
    ! Release memory
    !******************
    deallocate(fileName)
end subroutine ommosaic

!*******************************************************************
! Calculate size of output image in physical units and in
! pixels. Achieves this by casting the coordinates of the corner
! pixels into the frames of the first image.
!*******************************************************************
subroutine calculateOutputSize(nInput, input, output)
    integer, intent(in) :: nInput
    type (ImageT), intent(inout) :: input(nInput), output
   

    ! Local variables
    integer :: i, j, k, xmin, xmax, ymin, ymax
    real (double) :: az, el, axvector(3), vlocal(3), vsky(3),&
         & xpix, ypix, outaz, outel

    ! Initialise bounds (in pixels)
    xmin = 0 
    xmax = 0 
    ymin = 0 
    ymax = 0

    !*****************************************************************
    ! Create local -> celestial transformation matrix for output image
    !*****************************************************************
    axvector(1) = output%xrval * DegToRad
    axvector(2) = output%yrval * DegToRad
    axvector(3)= output%roll * DegToRad
    call sla_deuler('ZYX', axvector(1), -axvector(2), axvector(3), &
         & output%sky2img)

    !****************************************************************
    ! Loop through the input images, casting each of the four corners
    ! into the frame of the output image. This allows us to find the
    ! bounding box of the input images in the output frame.
    !****************************************************************
    do i = 1, nInput
       axvector(1) = input(i)%xrval * DegToRad
       axvector(2) = input(i)%yrval * DegToRad
       axvector(3) = input(i)%roll  * DegToRad
       call sla_deuler('ZYX', axvector(1), -axvector(2), axvector(3),&
            & input(i)%sky2img)

       ! Initialise bounding box
       input(i)%xmin = 0
       input(i)%xmax = 0
       input(i)%ymin = 0
       input(i)%ymax = 0


       ! NB. There's a half-pixel shift in here somewhere which needs to
       ! be sorted out (thanks FITS people. No, really, thanks).
       !
             
       ! Loop through each corner
       do j = 0, 1
          az = (dble(input(i)%dims(0) * j) - input(i)%xrpix + 0.5_double) * input(i)%xdelt


          do k = 0, 1
             el = (dble(input(i)%dims(1) * k) - input(i)%yrpix + 0.5_double) * input(i)%ydelt

             !******************************************************* 
             ! Transform corner from local frame to output frame. I'm
             ! doing this the long-handed (and stupid) way as it's
             ! easier to get ones head around.
             !******************************************************

             ! Convert spherical -> cartesian
             call sla_dcs2c(az * DegToRad, el * DegToRad, vlocal)

             ! Transform from local to celestial (cartesian)
             call sla_dimxv(input(i)%sky2img, vlocal, vsky)
             
             ! Transform into output frame
             call sla_dmxv(output%sky2img, vsky, vlocal)

             ! Convert to back spherical coordinates
             call sla_dcc2s(vlocal, outaz, outel)

             ! Convert into pixels in the output frame
             xpix = outaz * RadToDeg / output%xdelt
             ypix = outel * RadToDeg / output%ydelt

             ! Check against global bounds
             xmin = min(xmin, floor(xpix))
             xmax = max(xmax, ceiling(xpix))
             ymin = min(ymin, floor(ypix))
             ymax = max(ymax, ceiling(ypix))

             ! Update local bounds
             input(i)%xmin = min(input(i)%xmin, floor(xpix))
             input(i)%xmax = max(input(i)%xmax, ceiling(xpix))
             input(i)%ymin = min(input(i)%ymin, floor(ypix))
             input(i)%ymax = max(input(i)%ymax, ceiling(ypix))

          end do
       end do
    end do

    ! Update reference pixel position of output image
    output%xrpix = -xmin
    output%yrpix = -ymin
    

    ! Calculate size of image, allocated array
    output%dims(0) = (xmax - xmin + 1)
    output%dims(1) = (ymax - ymin + 1)

    allocate(output%image(0 : output%dims(0) - 1, 0 : output%dims(1) - 1))
    output%image = -1.0   ! Was 0.0

    allocate(output%weight(0 : output%dims(0) - 1, 0 : output%dims(1) - 1))
    output%weight = 0.0

    ! Translate bounds of all input images
    do i = 1, nInput
        input(i)%xmin = input(i)%xmin - xmin
        input(i)%xmax = input(i)%xmax - xmin
        input(i)%ymin = input(i)%ymin - ymin
        input(i)%ymax = input(i)%ymax - ymin
    end do
     
end subroutine calculateOutputSize
!********************************************************
! Read image and associated WCS keywords from a dataset
!
!*********************************************************
subroutine readImage(image, imagefile, obsDetails)
    type (ImageT), intent(out) :: image
    character (*), intent(in) :: imagefile
    type (OBSERVATIONDETAILS), intent(inout) :: obsDetails 
    type (OBSERVATIONDETAILS) :: newObsDetails 
    
    ! Local variables
    integer, dimension(:), pointer :: dims
    type (DataSetT) :: set
    type (ArrayT) :: arr
    integer :: k
    character(50) :: startObs
    character(50) :: endObs
    logical, save :: first = .true. 
    ! Open dataset, find image array
    image%filename = imagefile

    set = dataset(imagefile, read)

  

    call GetObservationDetails(set, newObsDetails)
    startObs = newObsDetails%startObs
    endObs = newObsDetails%endObs

    if (first) then
        first = .false.
        obsDetails = newObsDetails
    else 
        k = CompareDates(obsDetails%startObs, startObs)
        
        if (k .lt. 0) obsDetails%startObs = startObs

        k = CompareDates(obsDetails%endObs, endObs)
        
        if (k .gt. 0) obsDetails%endObs = endObs
    end if

 
    arr = array(set, 'PRIMARY', read)
    
    ! Check dimensions of image
    if(numberOfDimensions(arr) .ne. 2) &
         call fatal('badDims', 'Only 2-D images are supported')

    dims=>dimensions(arr)
    allocate(image%image(0 : dims(0) - 1, 0 : dims(1) - 1))
    allocate(image%weight(0 : dims(0) - 1, 0 : dims(1) - 1))
    
    
    image%dims = dims

    ! Read WCS keywords
    image%xrpix = real64Attribute(arr, 'CRPIX1')
    image%xrval = real64Attribute(arr, 'CRVAL1')
    image%xdelt = real64Attribute(arr, 'CDELT1')
    image%yrpix = real64Attribute(arr, 'CRPIX2')
    image%yrval = real64Attribute(arr, 'CRVAL2')
    image%ydelt = real64Attribute(arr, 'CDELT2')
   
    image%exposure = real32Attribute(set, "EXPOSURE")
    image%binAx1 = int32Attribute(set, "BINAX1")

  
    image%roll = 0.0


   ! Perform type conversion on image data
    image%datatype = datatype(arr)
    call readImageAsReal32(arr, image%image)



    !*******************************
    ! Read in the exposure image
    !*******************************
    if (hasBlock(set, "EXPOSURE")) then
        arr = array(set, 'EXPOSURE', read)
       
 
        call readImageAsReal32(arr, image%weight)
        
        where(image%weight .gt. 0.0)  ! NB was image%image 
            image%weight = 1.0
        elsewhere
            image%weight = 0.0
        end where   
         
    else
        where(image%image .ge. 0.0) 
            image%weight = 1.0
          
            
        elsewhere
            image%image = -1.0
            image%weight = 0.0
            
        end where
    end if
  
     

    ! Free the dataset
    call release(arr)
    call release(set)


end subroutine readImage

!********************************************
! Add an image to the output image
!
!********************************************
subroutine addImage(input, output, exposureImage)
    type (ImageT), intent(in) :: input
    type (ImageT), intent(inout) :: output
    real(single), intent(inout), dimension(:, :) :: exposureImage  
 
    ! Local variables
    integer :: i, j, xpix, ypix
    real (double) :: iaz, iel, oel, vlocal(3), vsky(3)
    real(double), dimension(:), allocatable :: oaz  
    !****************************************************************
    ! Loop over every pixel in the output image, transform from local
    ! coordinates to sky coordinates, transform from sky coordinates
    ! local coordinates in the input image, pick up pixel value and
    ! copy to output pixel
    !
    ! The matrix stuff here could be optimised quite a bit I think
    !***************************************************************

    allocate(oaz(1 :  output%dims(0)))
    forall (i = 1 :  output%dims(0)) oaz(i) = &
               ((dble(i) - output%xrpix) * output%xdelt) * DegToRad


    do j = 0, output%dims(1) - 1
        oel = ((dble(j + 1) - output%yrpix) * output%ydelt) * DegToRad

        do i = 0, output%dims(0) - 1
  

            !***********************************
            ! Convert spherical -> cartesian
            !***********************************
            call sla_dcs2c(oaz(i + 1), oel, vlocal)

            !****************************************************
            ! Transform from local to celestial (cartesian)
            ! sla_dimxv applies a 3d reverse rotation 
            ! sky2img is the rotation matrix, vlocal the
            ! vector to be rotated and vsky the resultant vector   
            !****************************************************
            call sla_dimxv(output%sky2img, vlocal, vsky)

            !********************************
            ! Transform into output frame
            ! sla_dmxv applies a 3-d rotation
            ! sky2img is the rotation matrix,
            ! vsky the vector to be rotated and
            ! vlocal the resultant vector
            !********************************
            call sla_dmxv(input%sky2img, vsky, vlocal)

            !*******************************************
            ! Convert to back spherical coordinates
            ! sla_dcc2s converts cartesian to spherical
            !*******************************************
            call sla_dcc2s(vlocal, iaz, iel)

            !*********************************************************
            ! Convert into pixels in the output frame
            ! NB - xdelt is the increment per pixel in the x-direction
            !      ydelt is the increment per pixel in the y-direction
            !      xrpix is the x-pixel reference point
            !      yrpix is the y-pixel reference point
            !**********************************************************
            xpix = nint(iaz * RadToDeg / input%xdelt + input%xrpix) - 1
            ypix = nint(iel * RadToDeg / input%ydelt + input%yrpix) - 1

            !******************************************
            ! Check against bounds of input image
            !******************************************
            if(xpix .gt. 0 .and. xpix .lt. input%dims(0) .and. &
                   ypix .gt. 0 .and. ypix .lt. input%dims(1)) then

                if (output%image(i, j) .lt. 0.0 .and. &
                    input%image(xpix, ypix) .ge. 0.0) output%image(i, j) = 0.0  ! NB was -1
                if (input%image(xpix, ypix) .gt. 0.0) then

                    output%image(i, j) = output%image(i, j) + &
                     & input%image(xpix, ypix)
                end if

                !************************************************
                ! Sum up the weighting image for the normalisation
                !************************************************
                output%weight(i, j) = output%weight(i, j) + input%weight(xpix, ypix)

                !************************************************
                ! Sum up the exposure image for the output file
                !************************************************
                if (input%weight(xpix, ypix) .gt. 0.0) &
                    exposureImage(i, j) = exposureImage(i, j) + input%exposure  
                
            end if

        end do
    end do
    deallocate(oaz)
end subroutine addImage
!**************************************************
! Write image to a new file
!**************************************************
subroutine writeImage(image, filename, obsDetails, exposureMap)
    type (imageT), intent(inout) :: image
    character (len=*), intent(in) :: filename
    type (OBSERVATIONDETAILS), intent(in) :: obsDetails
    logical, intent(in) :: exposureMap

    ! Local variables
    type (DataSetT) :: set
   
   
    ! Create new dataset
    set = dataSet(filename, CREATE)


    !*******************************************
    ! Add the fits header information
    !*******************************************
    call AddFitsKeyWords(set, obsDetails)


 
    !********************************
    ! Add the image array
    !********************************
    
    where(image%weight .eq. 0.0) 
         image%weight = -1.0
         image%image = -1.0
    end where



    call AddImageArray(set, "PRIMARY", image, image%image, .true.)


    !***************************************
    ! Add the exposure array (if required)
    !***************************************
    if (exposureMap) then
        call AddImageArray(set, "EXPOSURE", image, image%weight, .false.)
    end if
 

    call release(set)
    
end subroutine writeImage

!********************************************************************
! Read an image as Real32. Assumes pointer points to correctly sized
! array of doubles.
!*******************************************************************
subroutine readImageAsReal32(arr, image)   ! Was real 64
    type (ArrayT), intent(in) :: arr
    real (single), pointer, dimension(:,:) :: image

    ! Local variables
    real (single), pointer :: f32array(:,:)
    real (double), pointer :: f64array(:,:)
    integer (int16), pointer :: i16array(:,:)
    integer (int32), pointer :: i32array(:,:)
   

    ! Perform type conversion on image data
   

    select case (datatype(arr))
        case (integer16)
            i16array=>int16Array2Data(arr)
            image = i16array
        case (integer32)
            i32array=>int32Array2Data(arr)
            image = i32array
        case (real32)
            f32array=>real32Array2Data(arr)
            image = f32array
        case (real64)
            f64array=>real64Array2Data(arr)
            image = f64array
        case default
            call error('badType', 'Image is of unsupported data type')
    end select
    
end subroutine readImageAsReal32

!**************************************************
! CheckFilters
! Checks that the filters in the files are all
! identical
!**************************************************
subroutine CheckFilters(nosw, oswFile, obsDetails)
    integer, intent(in) :: nosw
    integer, parameter :: nFilt = 7
    character(*), intent(in), dimension(1 : nosw) :: oswFile
    Type (OBSERVATIONDETAILS), intent(inout) :: obsDetails
    character(20), allocatable, dimension(:) :: filter
    character(20) :: oswFilter
    character(100) :: string
    integer :: numFilters 
    integer :: i, j
    type (DataSetT) :: oswDataSet
    logical :: found

    allocate(filter(1 : nosw))
 

    numFilters = 1
    do i = 1 , nosw
        oswDataSet = DataSet(oswFile(i),read)
        oswFilter = trim(stringAttribute(oswDataSet, "FILTER"))
        
        if (i .eq. 1) then
            filter(1) = oswFilter
        else
            found = .false. 
            do j = 1, numFilters
        

                if (oswFilter .eq. filter(j)) then
                    found = .true.
                    exit
                end if
            end do
            if (.not. found) then
                numFilters = numFilters + 1 
                filter(numFilters) = oswFilter
            end if 

      
        end if    
        call release(oswDataSet)
   
 
    end do
  
    if (numFilters .gt. 1) then
       string = "("
       do i = 1, numFilters - 1
           string = trim(string) // trim(filter(i)) // ", "
       end do
       
       string = trim(string) // trim(filter(numFilters)) // ")"

       call warning("You are combining images from different filters!", trim(string))
       obsDetails%filter = "VARIOUS"
    end if

    deallocate(filter)

end subroutine CheckFilters
!********************************************************************
! Subroutine CheckWindows
! Purpose - To check to see if any image overlaps with another one.
! If so, outputs a warning message.
!*******************************************************************
subroutine CheckWindows(nosw, oswFile)
    integer, intent(in) :: nosw
    character(*), intent(in), dimension(1 : nosw) :: oswFile
    character(20), allocatable, dimension(:) :: ExposureNo
    character(20) :: exposureID
    character(100) :: string
    integer :: i, j
    type (DataSetT) :: oswDataSet
    
    allocate(exposureNo(1 : nosw))

    do i = 1, nosw

        oswDataSet = DataSet(oswFile(i),read)

        !****************************
        ! Get the exposure No
        !****************************
        exposureID = trim(stringAttribute(oswDataSet, "EXP_ID"))
        read(exposureID(11 : 13), *) exposureNo(i)
    
        call release(oswDataSet)

        do j = 1, i - 1
          

            if (oswFile(i) .eq. oswFile(j)) then     
                string = "(" // trim(oswFile(i)) // ")" 
                        call warning("You are combining identical images!", trim(string))
            else
                if (exposureNo(i) .eq. exposureNo(j)) then
                    string = "(" // trim(oswFile(i)) // ", " // trim(oswFile(j)) // ")"
                    call warning("You are combining images from the same exposure!", trim(string))
                end if
            end if
           
        end do
    end do
    
    deallocate(exposureNo)
    
end subroutine CheckWindows
!********************************************
! ConvertToLowResImage
! Converts an image from high-resolution
! to low-resolution
!********************************************
subroutine ConvertToLowResImage(image)

    type (imageT), intent(inout) :: image
    type (imageT) :: lowResImage
    integer :: i, j, i1, j1, k, l
    integer :: count

    if ((image%dims(0) / 2) * 2 .eq. image%dims(0)) then 
        lowResImage%dims(0) = image%dims(0) / 2 
    else
        lowResImage%dims(0) = image%dims(0) / 2 + 1 
    end if  

    if ((image%dims(1) / 2) * 2 .eq. image%dims(1)) then 
        lowResImage%dims(1) = image%dims(1) / 2 
    else
        lowResImage%dims(1) = image%dims(1) / 2 + 1 
    end if  


    !*************************************
    ! Create the new low-res image array
    !*************************************
    allocate(lowResImage%image(0 : lowResImage%dims(0) - 1, 0 : lowResImage%dims(1) - 1))
    lowResImage%image = 0.0
    allocate(lowResImage%weight(0 : lowResImage%dims(0) - 1, 0 : lowResImage%dims(1) - 1))
    lowResImage%weight = 0.0

    !******************************************************
    ! Need to co-add the four pixels of the high-res image
    ! to produce a low-res pixel
    !******************************************************

    do i = 0, image%dims(0) - 1, 2
        i1 = i / 2        
        do j = 0, image%dims(1) - 1, 2
            j1 = j / 2
            count = 0            
            do k = i, min(i + 1, image%dims(0) - 1) 
                do l = j, min(j + 1, image%dims(1) - 1)
                    lowResImage%image(i1, j1) = lowResImage%image(i1, j1) + image%image(k, l)
                    lowResImage%weight(i1, j1) = lowResImage%weight(i1, j1) + image%weight(k, l)
                    count = count + 1 
                end do
          
            end do 
            !********************************************************************
            ! Normalize this pixel by no of pixels (usually 4) used in averaging
            !********************************************************************
            lowResImage%weight(i1, j1) = lowResImage%weight(i1, j1) / real(count)

        end do
    end do
    
    image%dims(0) = lowResImage%dims(0) 
    image%dims(1) = lowResImage%dims(1)

    image%xdelt = image%xdelt * 2.0_double
   
    image%ydelt = image%ydelt * 2.0_double
     
    image%xrpix = image%xrpix * 0.5_double
    image%yrpix = image%yrpix * 0.5_double
    
    deallocate(image%image)
    deallocate(image%weight)

    allocate(image%image(image%dims(0), image%dims(1)))
    allocate(image%weight(image%dims(0), image%dims(1)))

    image%image = lowResImage%image
    image%weight = lowResImage%weight

    deallocate(lowResImage%image)
    deallocate(lowResImage%weight)

end subroutine ConvertToLowResImage
!****************************************
! Subroutine AddImageArray
!****************************************
subroutine AddImageArray(set, name, image, imageArray, remove)
    type (DataSetT), intent(in) :: set
 
    character(*), intent(in) :: name
    type(imageT), intent(in) :: image
    real(single), intent(in), dimension &
               (0 : image%dims(0) - 1, 0 : image%dims(1) - 1) :: imageArray
    logical, intent(in) :: remove
    type (ArrayT) :: arr
    real (single), dimension(:, :), pointer :: optr
    integer(int32) :: i, j, k   


    arr = addArray(set, name, REAL32, dimensions = image%dims)
    optr=>real32Array2Data(arr)
 
    !******************************************
    ! Set pixels outside image boundary to NULL
    ! Have to set all the pixels to NULL to 
    ! begin with then set positive pixels
    !******************************************


    k = 0
    do i = 0, image%dims(0) - 1
        do j = 0, image%dims(1) - 1
            call setToNull(arr, k)
           
            k = k + 1
        end do
    end do 
 

  
    do i = 0, image%dims(0) - 1
        do j = 0, image%dims(1) - 1
            if (imageArray(i, j) .ge. 0.0) then
                optr(i, j) = imageArray(i, j)
            end if
            
        end do
    end do 
 
   
    !***********************************************
    ! Remove gaps in image
    !**********************************************
    if(remove) call RemoveGaps(image, imageArray, optr)

    ! Add WCS keywords
  
    call setAttribute(arr, 'CTYPE1', 'RA---TAN',&
         & 'Coord. type: RA tangent plane projection')
    call setAttribute(arr, 'CRPIX1', image%xrpix, 'WCS reference pixel')
    call setAttribute(arr, 'CRVAL1', image%xrval,&
         & 'coord. at X axis ref. pixel')
    call setAttribute(arr, 'CDELT1', image%xdelt, 'WCS pixel size')
    call setAttribute(arr, 'CTYPE2', 'DEC--TAN',&
         & 'Coord. type: DEC tangent plane projection')
    call setAttribute(arr, 'CRPIX2', image%yrpix, 'WCS reference pixel')
    call setAttribute(arr, 'CRVAL2', image%yrval,&
         & 'coord. at Y axis ref. pixel')
    call setAttribute(arr, 'CDELT2', image%ydelt, 'WCS pixel size')

    ! Release dataset
    call release(arr)

end subroutine AddImageArray

!*******************************************************
! CompareDates - Compares the first date with the second.
! Returns -1 if the second is less than the first,
! 0 if they are equal and 1 if greater.
!*******************************************************
integer function CompareDates(dateString1, dateString2)
    character(*), intent(in) :: dateString1, dateString2
    integer :: time1(6), time2(6)        ! Stores the time from the date string
    character(1) char    ! Used for date and time extraction 
    integer i

    read(dateString1, '(i4, a1, i2, a1, i2, a1, i2, a1, i2, a1, i2)') &
         time1(1), char, time1(2), char, time1(3), char, time1(4), &
         char, time1(5), char, time1(6)

    read(dateString2, '(i4, a1, i2, a1, i2, a1, i2, a1, i2, a1, i2)') &
         time2(1), char, time2(2), char, time2(3), char, time2(4), &
         char, time2(5), char, time2(6)

    do i = 1, 6

        CompareDates = time2(i) - time1(i)
        if (CompareDates .ne. 0) return
 
    end do

    CompareDates = 0 ! Must be equal

end function CompareDates

!*************************************************************
! Subroutine GetObservationDetails
! Retrieves the observation details from a fits file
! for plottiong on the graph
!************************************************************
subroutine GetObservationDetails(dataSet, obsDetails)
    type(DataSetT), intent(in) :: dataSet
    type(OBSERVATIONDETAILS), intent(inout) :: obsDetails  


    call GetStringAttribute(dataSet, "TELESCOP", obsDetails%telescope)
    call GetStringAttribute(dataSet, "INSTRUME", obsDetails%instrument)
    call GetStringAttribute(dataSet, "OBSERVER", obsDetails%observer)
    call GetStringAttribute(dataSet, "OBJECT", obsDetails%object)


    call GetStringAttribute(dataSet, "FILTER", obsDetails%filter)
 
    call GetStringAttribute(dataSet, "OBS_ID",   obsDetails%obsID)
    call GetStringAttribute(dataSet, "EXP_ID",   obsDetails%expID)
    call GetStringAttribute(dataSet, "DATE-OBS", obsDetails%startObs)
    call GetStringAttribute(dataSet, "DATE-END", obsDetails%endObs)
    call GetRealAttribute  (dataSet, "EXPOSURE", obsDetails%exposure)

end subroutine GetObservationDetails

!***********************************************************
! Subroutine GetStringAttribute - 
!            Attempts to read a string attribute
!                        
!***********************************************************
subroutine GetStringAttribute(dataSet, attribute, string)
    type(DataSetT), intent(in) :: dataSet
    character(*), intent(in) :: attribute
    character(*), intent(inout) :: string

    if (.not. hasAttribute(dataSet, attribute)) then
        string = ""
    else
        string = trim(stringAttribute(dataSet, attribute))
    end if
end subroutine GetStringAttribute

!***********************************************************
! Subroutine GetRealAttribute - 
!         Attempts to read a real attribute
!                        from a fits file
!                      
!***********************************************************
subroutine GetRealAttribute(dataSet, attribute, number)
    type(DataSetT), intent(in) :: dataSet
    character(*), intent(in) :: attribute
    real, intent(inout) :: number

    if (.not. hasAttribute(dataSet, attribute)) then
        number= 0.0
    else
        number = real32Attribute(dataSet, attribute)
    end if
   
end subroutine GetRealAttribute

!********************************************************
! Subroutine AddFitsKeyWords
!*******************************************************
subroutine AddFitsKeyWords(set, obsDetails)
    type (DataSetT), intent(in) :: set
    type (OBSERVATIONDETAILS), intent(in) :: obsDetails

    !****************************************
    ! Add various keywords to set.
    !****************************************
    call setAttribute(set, "RADECSYS", "FK5", " ")
    call setAttribute(set, "EQUINOX", 2000, "", "Equinox of celestial coordinate system")
    call setAttribute(set, "EXPOSURE", obsDetails%exposure, "secs", "Exposure")
    call setAttribute(set, "DATE-OBS", obsDetails%startObs, "Start date of the first exposure")
    call setAttribute(set, "DATE-END", obsDetails%EndObs,   "End date of the last exposure")
    call setAttribute(set, "OBSERVER", obsDetails%observer, "Name of PI")
    call setAttribute(set, "OBJECT",   obsDetails%object, "Name of object")


    call setAttribute(set, "INSTRUME", obsDetails%instrument, "Instrument")
    call setAttribute(set, "FILTER", obsDetails%filter, "Filter")

 
!    call OAL_addCommonAttributes(set)


end subroutine AddFitsKeyWords





!***************************************************
! Subroutine RemoveGaps
!***************************************************
subroutine RemoveGaps(image, imageArray, optr)
    type(imageT), intent(in) :: image
    real(single), intent(in), dimension &
               (0 : image%dims(0) - 1, 0 : image%dims(1) - 1) :: imageArray
   
    real (single), dimension(:, :), pointer :: optr
    integer(int32) :: i, j, k, l, m  
      
    integer :: jStart = -1, jFinish = -1, nSize
    logical :: start
    integer, parameter :: maxSize = 100, width = 2
    real(single), pointer :: data(:) => null()
 


    allocate(data(1 : maxSize))

    do i = 3, image%dims(0) - 4
        jStart = -1
        start = .false.
        nSize = 0

        do j = 3, image%dims(1) - 4
            
            if (imageArray(i, j) .ge. 0.0) then
                start = .true.

                optr(i, j) = imageArray(i, j)
                if (jStart .gt. 0 .and. jFinish .ge. jStart) then ! Replace nan pixel value with median of surrounding pixels
                    if (jFinish - jStart .lt. 12) then
                        do k = jStart, jFinish
                            nSize = 0
                            lLoop : do l = max(0, k - width), min(k + width, image%dims(1) - 1)
                          
                                do m = max(0, i - width), min(i + width, image%dims(0) - 1)
                                    if(imageArray(m, l) .ge. 0.0) then
                                        nSize = nSize + 1
                                        data(nSize) = imageArray(m, l)
                                        if (nSize .eq. maxSize) exit lLoop
                                    end if
                                end do
                            end do lLoop
                            if (nSize .gt. 2) then  
                                optr(i, k) = MedianValue1(data, nSize)
                                 
                            else
                                optr(i, k) = (imageArray(i, jStart - 1) + imageArray(i, jFinish + 1)) / 2.0
                                
                            end if  
                        end do

                        jStart = -1
                        jFinish = -1    
                    else
                        jStart = -1
                        jFinish = -1
                       
                    end if
                end if 
            else
                if (jStart .lt. 0 .and. start) then
                    jStart = j
                    jFinish = j
                else
                    jFinish = j
                end if  

            end if
       
        end do


    end do 
    deallocate(data)
end subroutine RemoveGaps   




end module ommosaic_mod









